#!/usr/bin/python3

from typing import Dict, Any, Tuple
from configparser import ConfigParser
from datetime import datetime
import logging, json, copy

from bottle import route, run, response, request, HTTPError


logging.basicConfig(format=u"%(levelname)-8s [%(asctime)s] %(message)s", level=logging.DEBUG, filename=u"uzofdtest.log")
substitutions = {}
methods = {}
errors = {36870:"ERROR_RECEIPT_COUNT_ZERO",
        36871:"ERROR_RECEIPT_INDEX_OUT_OF_BOUNDS",
        36872:"ERROR_RECEIPT_NOT_FOUND",
        36873:"ERROR_DATA_SIZE_NOT_SUPPORTED",
        36874:"ERROR_RECEIPT_ALREADY_CLOSED",
        36875:"ERROR_RECEIPT_IS_NOT_CLOSED",
        36876:"ERROR_EXTRA_DATA_PASSED",
        36877:"ERROR_RECEIPT_FORMAT_INVALID",
        36878:"ERROR_RECEIPT_TOTAL_PRICE_OVERFLOW",
        36879:"ERROR_RECEIPT_TOTAL_PRICE_MISMATCH",
        36880:"ERROR_RECEIPT_PRICE_OVERFLOW",
        36881:"ERROR_RECEIPT_VAT_OVERFLOW",
        36882:"ERROR_RECEIPT_SEQ_OVERFLOW",
        36883:"ERROR_RECEIPT_CASH_OVERFLOW",
        36884:"ERROR_RECEIPT_CARD_OVERFLOW",
        36885:"ERROR_RECEIPT_PARTITION_INVALID",
        36886:"ERROR_RECEIPT_MEMORY_FULL",
        36888:"ERROR_RECEIPT_TIME_PAST",
        36889:"ERROR_RECEIPT_STORE_DAYS_LIMIT_EXCEEDED",
        36890:"ERROR_LAST_TRANSACTION_TIME_FORMAT_INVALID",
        36891:"ERROR_FIRST_RECEIPT_TRANSACTION_TIME_FORMAT_INVALID",
        36892:"ERROR_ACKNOWLEGE_WRONG_LENGTH",
        36893:"ERROR_ACKNOWLEGE_SIGNATURE_INVALID",
        36894:"ERROR_ACKNOWLEGE_TERMINAL_ID_MISMATCH",
        36895:"ERROR_ZREPORT_PARTITION_INVALID",
        36896:"ERROR_OPEN_CLOSE_ZREPORT_WRONG_LENGTH",
        36897:"ERROR_CLOSE_ZREPORT_TIME_PAST",
        36898:"ERROR_ZREPORT_SPACE_IS_FULL",
        36899:"ERROR_CURRENT_TIME_FORMAT_INVALID",
        36900:"ERROR_RECEIPT_TRANSACTION_TIME_FORMAT_INVALID",
        36902:"ERROR_ZREPORT_INDEX_OUT_OF_BOUNDS",
        36903:"ERROR_LOCK_CHALLENGE_INVALID",
        36904:"ERROR_LOCKED_FOREVER",
        36905:"ERROR_CONFIGURE_WRONG_LENGTH",
        36906:"ERROR_CONFIGURE_SIGNATURE_INVALID",
        36907:"ERROR_CURRENT_ZREPORT_IS_EMPTY",
        36908:"ERROR_RECEIPT_TOTAL_PRICE_ZERO",
        36909:"ERROR_ZREPORT_IS_NOT_OPEN",
        36910:"ERROR_ZREPORT_OPEN_TIME_FORMAT_INVALID",
        36911:"ERROR_SALE_REFUND_COUNT_OVERFLOW",
        36912:"ERROR_ZREPORT_IS_ALREADY_OPEN",
        36913:"ERROR_NOT_ENOUGH_CASH_FOR_REFUND",
        36914:"ERROR_NOT_ENOUGH_CARD_FOR_REFUND",
        36915:"ERROR_NOT_ENOUGH_VAT_FOR_REFUND",
        36916:"ERROR_OPEN_ZREPORT_TIME_PAST",
        36917:"ERROR_MAINTENANCE_REQUIRED",
        36918:"ERROR_CONFIGURE_UNLOCK_KEY_INVALID",
        65276:"bad refund info terminal id value",
        65533:"ERROR_CAN_NOT_SELECT_APPLET",
        65534:"ERROR_CAN_NOT_CONNECT_CARD"}


class Method(object):
    rules: Dict[int, Any]

    def __init__(self, name: str, defaultResponse):
        self.name = name
        self.defaultResponse = defaultResponse
        self.calls = 0
        self.rules = {}
        logging.info(u"Метод \"%s\" подготовлен" % name)

    def __call__(self, id: int) -> str:
        self.calls += 1
        logging.info(u"%iе обращение к методу \"%s\"" % (self.calls, self.name))
        response = copy.deepcopy(self.rules.get(self.calls, self.defaultResponse))

        if "_" in  response:
            logging.info(u"Статус ответа от сервера №%i" % response["_"])
            raise HTTPError(response["_"])

        if response:
            response["id"] = id
            if "result" in response:
                for key in response["result"]:
                    if key in substitutions and (response["result"][key] == "" or response["result"][key] == -1):
                        response["result"][key] = substitutions[key]
            strResponse = json.dumps(response)
        else:
            strResponse = ""

        logging.info(u"Ответ \"%s\"" % strResponse)
        return strResponse

    def addRule(self, callsStr: str, rule: Any):
        calls = callsStr.split(",")
        for call in calls:
            if call.isdigit():
                self.rules[int(call)] = rule
            else:
                parts = call.split("-")
                for number in range(int(parts[0]), int(parts[1]) + 1):
                    self.rules[number] = rule

def checkTime(time: datetime):
    currentTime = datetime.strptime(substitutions["CurrentTime"], "%Y-%m-%d %H:%M:%S")
    if strictTimeCheck:
        return (time - currentTime).total_seconds() > 0
    else:
        return (time - currentTime).total_seconds() >= 0

def makeError(id: int, code: int, message = "", log = True):
    if not message:
        message = errors.get(code, "Неизвестная ошибка! Всё плохо...")
    response = '{"jsonrpc":"2.0","id":%i,"error":{"code":%i,"message":"%s","data":"%s\\n%s"}}' % \
            (id, code, message, message, message)
    if log:
        logging.info(u"Ответ \"%s\"" % response)
    return response

@route("/rpc/api", method="POST")
def processRequest() -> str:
    response.content_type = 'application/json'
    try:
        body = json.loads(request.body.read())
    except:
        return makeError(0, -32700, u"Кривой JSON в запросе")
    logging.info(u"Обработка запроса \"%s\"" % body)
    method = body["method"].split(".")[1]
    data = body["params"]
    id = body["id"]
    if method == "SendSaleReceipt" or method == "SendRefundReceipt":
        if method ==  "SendRefundReceipt" and "RefundInfo" not in data["Receipt"].keys():
            return makeError(id, 65276)
        if useCheckLimit and substitutions["ReceiptCount"] == substitutions["ReceiptMaxCount"]:
            return makeError(id, 36886)
        time = datetime.strptime(data["Receipt"]["Time"], "%Y-%m-%d %H:%M:%S")
        if useTime and not checkTime(time):
            return makeError(id, 36888)
        substitutions["CurrentTime"] = data["Receipt"]["Time"]
        substitutions["DateTime"] = time.strftime("%Y%m%d%H%M%S")
        substitutions["ReceiptCount"] += 1
    elif method == "GetReceiptInfo":
        number = data["Number"]
        if useCheckLimit and (number > substitutions["ReceiptMaxCount"] or number < 0):
            return makeError(id, 36871)
        substitutions["Count"] = substitutions["ReceiptCount"]
        substitutions["Number"] = number
    elif method == "GetReceiptCount":
        substitutions["Count"] = substitutions["ReceiptCount"]
    elif method == "GetZReportInfo" or method == "GetZReportInfoByNumber":
        number = data["Number"]
        if useZReportLimit and (number > substitutions["ZReportMaxCount"] or number < 0):
            return makeError(id, 36902)
        substitutions["Count"] = substitutions["ZReportCount"]
        if method == "GetZReportInfo":
            substitutions["Number"] = substitutions["ZReportCount"] - number + 1
        else:
            substitutions["Number"] = number
    elif method == "OpenZReport":
        if useCheckLimit and substitutions["ZReportCount"] == substitutions["ZReportMaxCount"]:
            return makeError(id, 36898)
        time = datetime.strptime(data["Time"], "%Y-%m-%d %H:%M:%S")
        if useTime and not checkTime(time):
            return makeError(id, 36916)
        substitutions["CurrentTime"] = data["Time"]
        substitutions["DateTime"] = time.strftime("%Y%m%d%H%M%S")
    elif method == "CloseZReport":
        time = datetime.strptime(data["Time"], "%Y-%m-%d %H:%M:%S")
        if useTime and not checkTime(time):
            return makeError(id, 36897)
        substitutions["ZReportCount"] += 1
        substitutions["CurrentTime"] = data["Time"]
        substitutions["DateTime"] = time.strftime("%Y%m%d%H%M%S")
    elif method == "GetZReportCount":
        substitutions["Count"] = substitutions["ZReportCount"]

    return methods.get(method, methods["Unknown"])(id)

def parseRule(rule: str) -> Dict[str, Any]:
    if not rule:
        return {}
    if rule.startswith("status ") and rule.split(" ")[1].isdigit():
        result = {"_":int(rule.split(" ")[1])}
    else:
        if rule.startswith('{'):
            result = json.loads(rule)
        else:
            error = rule.split(" ")[0]
            if error.isdigit():
                code = int(error)
            else:
                code = 99999
                for key, value in errors.items():
                    if value == error:
                        code = key
            if rule.count(" ") < 1:
                result = json.loads(makeError(-1, code, log = False))
            else:
                result = {"error":{"code":code, "message":errors.get(code, "Неизвестная ошибка! Всё плохо..."),
                        "data":rule.split(" ", 1)[1]}}
        result["jsonrpc"] = "2.0"
    return result

def init() -> Tuple[str, int]:
    logging.info(u"Чтение настроек")
    config = ConfigParser()
    config.optionxform = str
    config.read("config.ini")
    global useTime
    global strictTimeCheck
    global useCheckLimit
    global useZReportLimit
    useTime = config.getboolean("Main", "useTime")
    strictTimeCheck = config.getboolean("Main", "strictTimeCheck")
    substitutions["CurrentTime"] = config.get("Main", "currentTime")
    substitutions["DateTime"] = substitutions["CurrentTime"].replace(" ", "")
    substitutions["DateTime"] = substitutions["DateTime"].replace("-", "")
    substitutions["DateTime"] = substitutions["DateTime"].replace(":", "")
    useCheckLimit = config.getboolean("Main", "useCheckLimit")
    substitutions["ReceiptCount"] = config.getint("Main", "checkCount")
    substitutions["ReceiptMaxCount"] = config.getint("Main", "checkMaxCount")
    useZReportLimit = config.getboolean("Main", "useZReportLimit")
    substitutions["ZReportCount"] = config.getint("Main", "zReportCount")
    substitutions["ZReportMaxCount"] = config.getint("Main", "zReportMaxCount")
    substitutions["TerminalID"] = config.get("Main", "terminalID")
    for key, value in config.items("DefaultResponse"):
        methods[key] = Method(key, parseRule(value))
    for key, value in config.items("ReActions"):
        parts = key.split(".")
        methods[parts[0]].addRule(parts[1], parseRule(value))
    return (config.get("Server", "host"), config.getint("Server", "port"))

def start() -> None:
    logging.info(u"Запускаем сервис")
    try:
        host, port = init()
        logging.info(u"Стартуем на %s:%d" % (host, port))
        run(host=host, port=port)
    except Exception as e:
        print(str(e))

if __name__ == "__main__":
    start()
